"""Integration-methods in scipy. This module contains their typing and can be
used for static type-checking."""
from typing import Callable, Tuple, Any, TypeVar, Optional

from scientific_plots.types_ import Matrix, Vector, Tensor

Input = TypeVar("Input", float, Matrix, Vector, Tensor,
                list[float])


def romberg(func: Callable[[Input], Input], x_min: float, x_max: float,
            tol: float = 1e-30,
            rtol: float = 1e-30,
            divmax: int = 15) -> float: ...


def quad(func: Callable[..., Input], xmin: float, xmax: float,
         limit: int = 50, epsabs: float = 1.49e-8, epsrel: float = 1.49e-8,
         args: Optional[tuple[float, ...]] = None)\
    -> Tuple[float, Any]: ...


def dblquad(func: Callable[[Input, Input], Input], a: float, b: float,
            gfun: Callable[[Input], Input],
            hfun: Callable[[Input], Input]) -> Tuple[float, Any]: ...


def simps(y: Input, x: Optional[Input] = None, dx: int = 1, axis: int = -1,
          even: str = "avg") -> float: ...


def trapz(y: Input, x: Optional[Input] = None, dx: float = 1, axis: int = -1)\
        -> float: ...



#!/usr/bin/env python3
"""
This module contains the settings for the various plots.
Plots can be created using the 'figure' deocorator from this module.
Multiple plots for various cases will be created and saved to
the hard drive
"""
from __future__ import annotations

import csv
import locale
from contextlib import contextmanager
from copy import copy, deepcopy
from functools import wraps
from typing import (
    Generator, Optional, Union, Callable, Any, overload)
from pathlib import Path
from warnings import warn, catch_warnings, simplefilter
from textwrap import dedent

import mpl_toolkits
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.pyplot import Axes
from matplotlib import colors
from cycler import cycler
import numpy as np

from .utilities import translate
from .types_ import Vector

mpl.use("Agg")
plt.rcParams["axes.unicode_minus"] = False

SPINE_COLOR = "black"
FIGSIZE = (3.15, 2.35)
FIGSIZE_SLIM = (3.15, 2.1)
FIGSIZE_SMALL = (2.2, 2.1)
_savefig = copy(plt.savefig)  # backup the old save-function


def linestyles() -> Generator[str, None, None]:
    """get the line-stiles as an iterator"""
    yield "-"
    yield "dotted"
    yield "--"
    yield "-."


rwth_colorlist: list[tuple[int, int, int]] = [(0, 84, 159), (246, 168, 0),
                                              (161, 16, 53), (0, 97, 101)]
rwth_cmap = colors.ListedColormap(rwth_colorlist, name="rwth_list")
mpl.colormaps.register(rwth_cmap)

rwth_hex_colors = ["#00549F", "#F6A800", "#A11035", "#006165",
                   "#57AB27", "#E30066"]

rwth_cycle = (
    cycler(color=rwth_hex_colors)
    + cycler(linestyle=["-", "--", "-.", "dotted",
                        (0, (3, 1, 1, 1, 1, 1)),
                        (0, (3, 5, 1, 5))]))

rwth_gradient: dict[str, tuple[tuple[float, float, float],
                               tuple[float, float, float]]] = {
    "red": ((0.0, 0.0, 0.0), (1.0, 142 / 255, 142 / 255)),
    "green": ((0.0, 84 / 255.0, 84 / 255), (1.0, 186 / 255, 186 / 255)),
    "blue": ((0.0, 159 / 255, 159 / 255), (1.0, 229 / 255, 229 / 255)),
}


def make_colormap(seq: list[tuple[tuple[Optional[float], ...],
                                  float,
                                  tuple[Optional[float], ...]]],
                  name: str = "rwth_gradient")\
        -> colors.LinearSegmentedColormap:
    """Return a LinearSegmentedColormap
    seq: a sequence of floats and RGB-tuples. The floats should be increasing
    and in the interval (0,1).
    """
    cdict: dict[str, list[tuple[float,
                                Optional[float],
                                Optional[float]
                                ]
                          ]] =\
        {"red": [], "green": [], "blue": []}
    for item in seq:
        red_1, green_1, blue_1 = item[0]
        red_2, green_2, blue_2 = item[2]

        cdict["red"].append((item[1], red_1, red_2))
        cdict["green"].append((item[1], green_1, green_2))
        cdict["blue"].append((item[1], blue_1, blue_2))
    return colors.LinearSegmentedColormap(name, cdict)


def partial_rgb(*x: float) -> tuple[float, ...]:
    """return the rgb value as a fraction of 1"""
    return tuple(v / 255.0 for v in x)


hks_44 = partial_rgb(0.0, 84.0, 159.0)
hks_44_75 = partial_rgb(64.0, 127.0, 183.0)
rwth_orange = partial_rgb(246.0, 168.0, 0.0)
rwth_orange_75 = partial_rgb(250.0, 190.0, 80.0)
rwth_gelb = partial_rgb(255.0, 237.0, 0.0)
rwth_magenta = partial_rgb(227.0, 0.0, 102.0)
rwth_bordeux = partial_rgb(161.0, 16.0, 53.0)


rwth_gradient_map = make_colormap(
    [
        ((None, None, None), 0., hks_44),
        (hks_44_75, 0.33, hks_44_75),
        (rwth_orange_75, 0.66, rwth_orange),
        (rwth_bordeux, 1., (None, None, None))
    ]
)
mpl.colormaps.register(rwth_gradient_map)


def _germanify(ax: Axes, reverse: bool = False) -> None:
    """
    translate a figure from english to german.
    The direction can be reversed, if reverse it set to True
    Use the decorator instead
    """

    for axi in ax.figure.axes:
        try:
            axi.ticklabel_format(
                useLocale=True)
        except AttributeError:
            pass
        items = [
            axi.xaxis.label,
            axi.yaxis.label,
            *axi.get_xticklabels(),
            *axi.get_yticklabels(),
        ]
        try:
            if axi.zaxis is not None:
                items.append(axi.zaxis.label)
                items += [*axi.get_zticklabels()]
        except AttributeError:
            pass
        if axi.get_legend():
            items += [*axi.get_legend().texts]
        for item in items:
            item.set_text(translate(item.get_text(),
                                    reverse=reverse))
    try:
        plt.tight_layout()
    except IndexError:
        pass


@contextmanager
def germanify(ax: Axes,
              reverse: bool = False) -> Generator[None, None, None]:
    """
    Translate the plot to german and reverse
    the translation in the other direction. If reverse is set to false, no
    reversal of the translation will be applied.
    """
    old_locale = locale.getlocale(locale.LC_NUMERIC)
    try:
        try:
            locale.setlocale(locale.LC_ALL, "de_DE")
            locale.setlocale(locale.LC_NUMERIC, "de_DE")
        except locale.Error:
            # locale not available
            pass
        plt.rcParams["axes.formatter.use_locale"] = True
        _germanify(ax)
        yield
    except Exception as e:
        print("Translation of the plot has failed")
        print(e)
        raise
    finally:
        try:
            locale.setlocale(locale.LC_ALL, old_locale)
            locale.setlocale(locale.LC_ALL, old_locale)
        except locale.Error:
            pass
        plt.rcParams["axes.formatter.use_locale"] = False
        if reverse:
            _germanify(ax, reverse=True)


def data_plot(filename: Union[str, Path]) -> None:
    """
    Write the data, which is to be plotted, into a txt-file in csv-format.
    """
    # pylint: disable=W0613
    if isinstance(filename, str):
        file_ = Path(filename)
    else:
        file_ = filename
    file_ = file_.parent / (file_.stem + ".csv")
    ax = plt.gca()
    try:
        with open(file_, "w", encoding="utf-8", newline="") as data_file:
            writer = csv.writer(data_file)
            for line in ax.get_lines():
                writer.writerow(
                    [line.get_label(), ax.get_ylabel(), ax.get_xlabel()])
                writer.writerow(line.get_xdata())
                writer.writerow(line.get_ydata())
    except PermissionError as e:
        print(f"Data-file could not be written for {filename}.")
        print(e)


def read_data_plot(filename: Union[str, Path])\
        -> dict[str, tuple[Vector, Vector]]:
    """Read and parse the csv-data-files, which have been generated by the
    'data_plot'-function."""
    data: dict[str, tuple[Vector, Vector]] = {}
    with open(filename, "r", newline="", encoding="utf-8") as file_:
        reader = csv.reader(file_)
        title: str
        x_data: Vector
        for i, row in enumerate(reader):
            if i % 3 == 0:
                title = row[0]
            elif i % 3 == 1:
                x_data = np.array(row, dtype=float)
            else:
                y_data: Vector
                y_data = np.array(row, dtype=float)
                data[title] = (x_data, y_data)
    return data


@contextmanager
def presentation_figure(figsize: tuple[float, float] = (4, 3)) ->\
        Generator[Axes, None, None]:
    """context manager to open an close the file.
    default seaborn-like plot"""
    fig, ax = plt.subplots(figsize=figsize)
    mpl.rcParams["text.latex.preamble"] = [
        r"\usepackage{helvet}",  # set the normal font here
        r"\usepackage{sansmath}",  # load up the sansmath so that math
        # -> helvet
        r"\sansmath",  # <- tricky! -- gotta actually tell tex to use!
    ]
    mpl.rc("font", family="sans-serif")
    mpl.rc("text", usetex=False)
    font = {"size": 30}

    mpl.rc("font", **font)
    plt.set_cmap("rwth_list")
    try:
        yield ax
    except Exception as e:
        print("creation of plot failed")
        print(e)
        raise
    finally:
        plt.close(fig)
        plt.close("all")
        mpl.rcParams.update(mpl.rcParamsDefault)
        plt.style.use("default")


old_save = plt.savefig


def alternative_save(
        filename: Path,
        dpi: Optional[int] = None,
        bbox_inches: Optional[Union[str, tuple[float, float]]] = None,
        figsize: tuple[float, float] = FIGSIZE,
        subfolder: str = "small") -> None:
    """
    Create additional saves of the given figsize and save these new figures
    into subfolder of given names. This function can be used to create
    additional plots of different sizes without a large overhead.
    """
    fig = deepcopy(plt.gcf())
    fig.set_size_inches(*figsize)
    with catch_warnings(record=True) as warning:
        simplefilter("always")
        fig.tight_layout()
        if warning:
            if issubclass(warning[-1].category, UserWarning):
                plt.close(fig)
                return
    folder = filename.parent / subfolder
    folder.mkdir(exist_ok=True)
    try:
        fig.savefig(
            folder
            / filename.name, dpi=dpi, bbox_inches=bbox_inches)
    except PermissionError:
        fig.savefig(
            folder
            / (filename.stem + "_" + filename.suffix),
            dpi=dpi, bbox_inches=bbox_inches)
    plt.close(fig)


def try_save(filename: Path,
             dpi: Optional[int] = None,
             bbox_inches: Optional[Union[str, tuple[float, float]]] = None, *,
             small: bool = False,
             slim: bool = False) -> None:
    """Try to save the current figure to the given path, if it is not possible,
    try to save it under a different name.
    If small is set to true, also create
    a smaller version of the given plot.
    If slim is set to true, a slightly slimmer version
    of the plot is created."""

    def alternative_save(
            figsize: tuple[float, float] = FIGSIZE,
            subfolder: str = "small") -> None:
        """
        Create additional saves of the given figsize and save these new figures
        into subfolder of given names. This function can be used to create
        additional plots of different sizes without a large overhead.
        """
        fig = deepcopy(plt.gcf())
        fig.set_size_inches(*figsize)
        with catch_warnings(record=True) as warning:
            simplefilter("always")
            fig.tight_layout()
            if warning:
                if issubclass(warning[-1].category, UserWarning):
                    plt.close(fig)
                    return
        folder = filename.parent / subfolder
        folder.mkdir(exist_ok=True)
        try:
            fig.savefig(
                folder
                / filename.name, dpi=dpi, bbox_inches=bbox_inches)
        except PermissionError:
            fig.savefig(
                folder
                / (filename.stem + "_" + filename.suffix),
                dpi=dpi, bbox_inches=bbox_inches)
        plt.close(fig)

    try:
        old_save(filename, dpi=dpi, bbox_inches=bbox_inches)
    except PermissionError:
        old_save(filename.parent / (filename.stem + "_" + filename.suffix),
                 dpi=dpi, bbox_inches=bbox_inches)

    if small:
        alternative_save(
            figsize=FIGSIZE_SMALL,
            subfolder="small")

    if slim:
        alternative_save(
            figsize=FIGSIZE_SLIM,
            subfolder="slim")


def new_save_simple(subfolder: Union[str, Path] = "", suffix: str = "", *,
                    german: bool = False, png: bool = True,
                    pdf: bool = True, small: bool = False,
                    slim: bool = False)\
        -> Callable[..., None]:
    """
    Return a new save function, which saves the file to a new given name in pdf
    format, and also creates a png version.
    If the argument "german" is set to true, also create German language
    version of the plots.
    """

    @wraps(old_save)
    def savefig_(filename: Union[Path, str],
                 dpi: Optional[int] = None,
                 bbox_inches: Optional[
                     Union[tuple[float, float], str]] = None) -> None:
        """Save the plot to this location as pdf and png."""
        if isinstance(filename, str):
            filename = Path(filename)
        if filename.parent == Path("."):
            warn(
                f"The filename {filename} in 'savefig' does "
                f"not contain a subfolder (i.e. 'subfolder/{filename})! "
                "Many files might be created onto the top level.")

        if subfolder:
            (filename.parent / subfolder).mkdir(exist_ok=True)
            new_path_pdf = filename.parent / subfolder / (
                filename.stem + suffix + ".pdf")
            new_path_png = filename.parent / subfolder / (
                filename.stem + suffix + ".png")
        else:
            new_path_pdf = filename.parent / (
                filename.stem + suffix + ".pdf")
            new_path_png = filename.parent / (
                filename.stem + suffix + ".png")

        # save the data
        data_path = filename.parent / (
            filename.stem + ".dat")

        if not data_path.exists():
            data_plot(data_path)

        try:
            plt.tight_layout()
        except IndexError:
            pass
        # save the figure
        if pdf:
            try_save(new_path_pdf, bbox_inches=bbox_inches,
                     small=small, slim=slim)
        if png:
            try_save(new_path_png, bbox_inches=bbox_inches,
                     dpi=dpi, small=small, slim=slim)

        if german:
            with germanify(plt.gca()):
                if pdf:
                    try_save(
                        new_path_pdf.parent
                        / (new_path_pdf.stem + "_german.pdf"),
                        bbox_inches=bbox_inches, small=small,
                        slim=slim)
                if png:
                    try_save(
                        new_path_png.parent
                        / (new_path_png.stem + "_german.png"),
                        bbox_inches=bbox_inches, dpi=dpi, small=small,
                        slim=slim)

    return savefig_


def presentation_settings() -> None:
    """Change the settings of rcParams for presentations."""
    # increase size
    fig = plt.gcf()
    fig.set_size_inches(8, 6)
    mpl.rcParams["font.size"] = 24
    mpl.rcParams["axes.titlesize"] = 24
    mpl.rcParams["axes.labelsize"] = 24
    # mpl.rcParams["axes.location"] = "left"
    mpl.rcParams["lines.linewidth"] = 3
    mpl.rcParams["lines.markersize"] = 10
    mpl.rcParams["xtick.labelsize"] = 18
    mpl.rcParams["ytick.labelsize"] = 18
    mpl.rcParams["figure.figsize"] = (10, 6)
    mpl.rcParams["figure.titlesize"] = 24

    mpl.rcParams["font.family"] = "sans-serif"


def set_rwth_colors(three_d: bool = False) -> None:
    """Apply the RWTH CD colors to matplotlib."""
    mpl.rcParams["text.usetex"] = False
    mpl.rcParams["axes.prop_cycle"] = rwth_cycle
    if three_d:
        plt.set_cmap("rwth_gradient")
    else:
        plt.set_cmap("rwth_list")


def set_serif() -> None:
    """Set the plot to use a style with serifs."""
    mpl.rcParams["font.family"] = "serif"
    mpl.rcParams["font.serif"] = [
        "cmr10", "stix", "Times New Roman"]
    mpl.rcParams["mathtext.fontset"] = "cm"


def set_sans_serif() -> None:
    """Set matplotlib to use a sans-serif font."""
    mpl.rcParams["font.family"] = "sans-serif"
    mpl.rcParams["font.sans-serif"] = [
        "Arial", "Helvetica", "DejaVu Sans"]


class ThreeDPlotException(Exception):
    """This exception is called when a 3D plot is drawn. This is used to exit
    the plotting function with the science-style."""


class FallBackException(Exception):
    """This is excaption is thrown when the fallback-style is selected.
    Only for debug purposes."""


def check_3d(three_d: bool) -> None:
    """This function checks if the current plot is a 3d plot. In that case, an
    exception is thrown, which can be used to stop the creation of the default
    plot."""
    if three_d:
        raise ThreeDPlotException
    if isinstance(plt.gca(), mpl_toolkits.mplot3d.axes3d.Axes3D):
        raise ThreeDPlotException


PlotFunction = Callable[..., None]


@overload
def apply_styles(plot_function: PlotFunction, *,
                 three_d: bool = False,
                 _fallback: bool = False) -> PlotFunction:
    ...


@overload
def apply_styles(plot_function: None, *, three_d: bool = False,
                 _fallback: bool = False)\
        -> Callable[[PlotFunction], PlotFunction]:
    ...


@overload
def apply_styles(*, three_d: bool = False,
                 _fallback: bool = False)\
        -> Callable[[PlotFunction], PlotFunction]:
    ...


def apply_styles(plot_function: Optional[PlotFunction] = None, *,
                 three_d: bool = False, _fallback: bool = False)\
        -> Union[Callable[[PlotFunction], PlotFunction], PlotFunction]:
    """
    Apply the newly defined styles to a function, which creates a plot.
    The new plots are saved into different subdirectories and multiple
    variants of every plot will be created.

    Arguments
    --------
    three_d: Create a use this option for 3D-plots
    fallback: switch directly to the fallback-style (for debug)
    """
    # pylint: disable=too-many-statements

    def _decorator(_plot_function: PlotFunction) -> PlotFunction:
        """This is the  actual decorator. Thus, the outer function
        'apply_styles' is actually a decorator-factory."""

        @wraps(_plot_function)
        def new_plot_function(*args: Any, **kwargs: Any) -> None:
            """
            New plotting function, with applied styles.
            """
            # default plot
            plt.set_cmap("rwth_list")
            plt.savefig = new_save_simple(png=False)
            _plot_function(*args, **kwargs)

            errors = (OSError, FileNotFoundError, ThreeDPlotException,
                      FallBackException)

            def journal() -> None:
                """Create a plot for journals."""
                set_rwth_colors(three_d)
                set_serif()
                plt.savefig = new_save_simple("journal", png=False,
                                              small=not three_d)
                _plot_function(*args, **kwargs)
                plt.close("all")

            def sans_serif() -> None:
                """
                Create a plot for journals with sans-serif-fonts.
                """
                set_rwth_colors(three_d)
                set_sans_serif()
                plt.savefig = new_save_simple("sans_serif", german=True,
                                              small=not three_d)
                _plot_function(*args, **kwargs)
                plt.close("all")

            def grayscale() -> None:
                """
                Create a plot in grayscales for disserations.
                """
                mpl.rcParams["text.usetex"] = False
                set_serif()
                if three_d:
                    plt.set_cmap("Greys")
                    new_kwargs = copy(kwargs)
                    new_kwargs["colorscheme"] = "Greys"
                else:
                    new_kwargs = kwargs
                plt.savefig = new_save_simple("grayscale", png=False,
                                              small=not three_d,
                                              slim=not three_d)
                _plot_function(*args, **new_kwargs)
                plt.close("all")

            def presentation() -> None:
                """
                Create a plot for presentations.
                """
                if three_d:
                    new_kwargs = copy(kwargs)
                    new_kwargs["figsize"] = (9, 7)
                    new_kwargs["labelpad"] = 20
                    new_kwargs["nbins"] = 5
                else:
                    new_kwargs = kwargs
                set_rwth_colors(three_d)
                presentation_settings()
                set_sans_serif()
                plt.savefig = new_save_simple("presentation",
                                              german=True, pdf=False)
                _plot_function(*args, **new_kwargs)
                plt.close("all")

            try:
                plt.close("all")

                check_3d(three_d)
                if _fallback:
                    raise FallBackException

                plt.close("all")

                # journal
                with plt.style.context(["science", "ieee"]):
                    journal()

                # sans-serif
                with plt.style.context(["science", "ieee", "nature"]):
                    sans_serif()

                # grayscale
                with plt.style.context(["science", "ieee", "grayscale"]):
                    grayscale()

                # presentation
                with plt.style.context(["science", "ieee"]):
                    presentation()

            except errors:
                if not three_d:
                    warn(dedent(""""Could not found style 'science'.
                                The package was probably installed incorrectly.
                                Using a fallback-style."""), ImportWarning)

                plt.close("all")
                # journal
                with plt.style.context("fast"):
                    if not three_d:
                        mpl.rcParams["figure.figsize"] = FIGSIZE
                        mpl.rcParams["font.size"] = 8
                    journal()

                # sans-serif
                with plt.style.context("fast"):
                    if not three_d:
                        mpl.rcParams["figure.figsize"] = FIGSIZE
                        mpl.rcParams["font.size"] = 8
                    sans_serif()

                # grayscale
                with plt.style.context("grayscale"):
                    if not three_d:
                        mpl.rcParams["figure.figsize"] = FIGSIZE
                        mpl.rcParams["font.size"] = 8
                    grayscale()

                # presentation
                with plt.style.context("fast"):
                    presentation()

            plt.savefig = old_save

        return new_plot_function

    if plot_function is not None:
        return _decorator(plot_function)

    return _decorator



#!/usr/bin/env python
"""
This are the typing-stubs for "matplotlib.pyplot".
They are needed for a testing using mypy.
"""
from __future__ import annotations
from typing import (
    Optional, Tuple, Iterable, Union, TypeVar, List, TypedDict, Any,
    overload)
from pathlib import Path
from contextlib import contextmanager
from collections.abc import Generator

from ..ticker import Locator
from ..figure import Figure
from .. import colors, RCParams

from scientific_plots.types_ import Vector, Matrix


In = TypeVar("In", List[float], Tuple[float],
             Vector, List[Any])

In2D = TypeVar("In2D", list[list[float]], list[Vector], tuple[Vector],
               Matrix)


def set_cmap(color: Union[colors.ListedColormap,
                          str,
                          colors.LinearSegmentedColormap]) -> None: ...


def savefig(
    filename: Union[Path, str],
    dpi: Optional[int] = None,
    bbox_inches: Optional[Union[tuple[float, float], str]] = None)\
        -> None: ...


def tight_layout() -> None: ...


class Label:
    def set_fontname(self, fontname: str) -> None: ...

    def set_fontsize(self, fontsize: int) -> None: ...

    def set_text(self, text: str) -> None: ...

    def get_text(self) -> str: ...

    def set_color(self, color: str) -> None: ...


class Line:
    def set_linewidth(self, width: int) -> None: ...

    def set_label(self, label: str) -> None: ...

    def get_label(self) -> Label: ...

    def __add__(self, other: Line) -> list[Line]: ...

    def get_xdata(self) -> Vector: ...

    def get_ydata(self) -> Vector: ...

    def get_color(self) -> str: ...


class Legend:
    texts: Tuple[Label, ...]


class Pane:
    def set_alpha(
        self,
        input_: float) -> None: ...


class Axis:
    label: Label
    labelpad: float
    pane: Pane

    def set_ticks_position(self, position: str) -> None: ...

    def set_tick_params(self, pad: Optional[float] = None,
                        direction: str = "", color: str = "")\
        -> None: ...

    def set_major_formatter(self, format_str: str) -> None: ...

    def set_major_locator(self, locator: Locator) -> None: ...

    def clear(self) -> None: ...

    def set_label_text(self, label: str) -> None: ...

    def get_label(self) -> str: ...


class Spine:
    """Spine-class of axes."""

    def set_color(self, color: str) -> None: ...

    def set_linewidth(self, width: float) -> None: ...

    def set_visible(self, visible: bool) -> None: ...

    def set_position(self, value: tuple[str, float]) -> None: ...


class Patch:
    """Patch class contained in axes."""

    def set_visible(self, value: bool) -> None: ...


class Axes:
    xaxis: Axis
    yaxis: Axis
    zaxis: Optional[Axis]
    axes: Iterable[Axis]
    dist: float = 10

    spines: dict[str, Spine]

    patch: Patch

    figure: Figure

    def set_yticks(self, ticks: list[float]) -> None: ...

    def set_yticklabels(self, labels: list[str]) -> None: ...

    def get_xticklabels(self) -> Iterable[Label]: ...

    def get_yticklabels(self) -> Iterable[Label]: ...

    def get_zticklabels(self) -> Iterable[Label]: ...

    def get_lines(self) -> Iterable[Line]: ...

    def get_legend(self) -> Legend: ...

    def set_xlabel(self, label: str,
                   linespacing: Optional[float] = None,
                   rotation: Optional[float] = None,
                   labelpad: Optional[float] = None,
                   color: Optional[str] = None) -> None: ...

    def set_ylabel(self, label: str,
                   linespacing: Optional[float] = None,
                   rotation: Optional[float] = None,
                   labelpad: Optional[float] = None,
                   color: Optional[str] = None) -> None: ...

    def set_zlabel(self, label: str,
                   linespacing: Optional[float] = None,
                   rotation: Optional[float] = None,
                   labelpad: Optional[float] = None,
                   color: Optional[str] = None) -> None: ...

    def set_xlim(self, min_: float, max_: float) -> None: ...

    def set_ylim(self, min_: float, max_: float) -> None: ...

    def set_zlim(self, min_: float, max_: float) -> None: ...

    def set_xscale(self, scale: str) -> None: ...

    def set_yscale(self, scale: str) -> None: ...

    def set_zscale(self, scale: str) -> None: ...

    def plot_surface(self, X: In2D, Y: In2D, Z: In2D, *,
                     cmap: str = "", alpha: Optional[float] = None) -> None: ...

    def scatter(self, X: In, Y: In, Z: In,
                cmap: Union[str, colors.ColorMap] = "jet") -> None: ...

    def plot_trisurf(self, X: In, Y: In, Z: In,
                     cmap: Union[str, colors.ColorMap] = "jet")\
        -> None: ...

    @overload
    def plot(self, X: In2D, Y: In2D, Z: In2D,
             fmt: Optional[str] = None, *,
             label: Optional[str] = None,
             color: Optional[str] = None, linestyle: Optional[str] = None,
             alpha: Optional[float] = None)\
        -> list[Line]: ...

    @overload
    def plot(self, X: In, Y: In, fmt: Optional[str] = None, *,
             label: Optional[str] = None,
             color: Optional[str] = None, linestyle: Optional[str] = None,
             alpha: Optional[float] = None)\
        -> list[Line]: ...

    def twinx(self) -> Axes: ...

    def get_ylabel(self) -> str: ...

    def get_xlabel(self) -> str: ...

    def set_frame_on(self, value: bool) -> str: ...

    def tick_params(self, axis: str = "x",
                    colors: str = "") -> None: ...

    def ticklabel_format(
        self,
        useLocale: bool = False, useMathText: bool = False) -> None: ...

    def set_box_aspect(
        self,
        aspect: Optional[str] = None, zoom: float = 1.)\
        -> None: ...

    def legend(
        self,
        handles: Optional[Union[list[Line], list[str]]] = None,
        labels: Optional[list[str]] = None) -> None: ...

    def set(
        self, **kwargs: Any) -> None: ...

    def set_title(self, label: str, loc: Optional[str] = None) -> None: ...

    def grid(self, visible: Optional[bool] = None, which: str = "major")\
        -> None: ...

    def hist(
        self, x: Union[Vector, Matrix],
        bins: Optional[Union[int, Vector, str]] = None,
        range: Optional[tuple[float, float]] = None,
        density: bool = False, weights: Optional[Vector] = None,
        cumulative: bool = False, bottom: Optional[float] = None,
        histtype: str = "bar", align: str = "mid",
        orientation: str = "vertical", log: bool = False,
        color: Optional[str] = None, label: Optional[str] = None,
        stacked: bool = False) -> None: ...


def figure(figsize: Optional[Tuple[float, float]] = None) -> Figure: ...


def close(fig: Union[Figure, str, None] = None) -> None: ...


def gcf() -> Figure: ...


def gca() -> Axes: ...


def xscale(scale: str) -> None: ...


def yscale(scale: str) -> None: ...


def plot(X: In, Y: In, label: Optional[str] = None,
         linestyle: str = "default",
         color: Optional[str] = None) -> None: ...


def xlim(min_: float, max_: float) -> None: ...


def ylim(min_: float, max_: float) -> None: ...


def xlabel(label: str) -> None: ...


def ylabel(label: str) -> None: ...


def legend(lines: Optional[list[Line]] = None,
           labels: Optional[list[Label]] = None) -> None: ...


def xticks(ticks: list[float],
           labels: Optional[list[str]] = None) -> None: ...


def yticks(ticks: list[float],
           labels: Optional[list[str]] = None) -> None: ...


class Style:
    """Styling of mpl."""

    def use(self, style: str) -> None: ...

    @contextmanager
    def context(self, style: Union[str, list[str]])\
        -> Generator[None, None, None]: ...


style: Style


def subplots(
    figsize: Tuple[float, float] = (10, 5)) -> Tuple[Figure, Axes]: ...


def show() -> None: ...


def ion() -> None: ...


def pause(t: float) -> None: ...


def clf() -> None: ...


def imshow(frame: int, cmap: Union[colors.ColorMap, str] = "default")\
        -> None: ...


def locator_params(
    *,
    nbins: int = 1,
    axis: str = "x") -> None: ...


rcParams: RCParams


class _ArrowProps(TypedDict):
    """Properties of an arrow."""
    arrowstyle: str
    shrinkA: float
    shrinkB: float


def annotate(
    text: str = "",
    s: str = "",
    xy: Optional[tuple[float, float]] = None,
    xytext: Optional[tuple[float, float]] = None,
    arrowprops: Optional[_ArrowProps] = None) -> None: ...


def text(x: float, y: float, text: str) -> None: ...

def axes(arg: Optional[tuple[str, str, str, str]] = None,
         label: str = "", projection: Optional[str] = None,
         polar: bool = False, sharex: Optional[Axes] = None,
         sharey: Optional[Axes] = None) -> Axes: ...


def subplot(
    *args: Any,
    label: str = "", projection: Optional[str] = None,
    polar: bool = False, sharex: Optional[Axes] = None,
    sharey: Optional[Axes] = None) -> Axes: ...



"""This is a stub-file for the numba-module for numeric python calculations. It
contains the typing for the used functions and functionals. This allows for a
type-checking using mypy."""

from typing import (
    Callable, TypeVar, Iterable, Iterator, Optional, Any)
from contextlib import contextmanager

from functools import wraps


Out = TypeVar("Out")


def jit(parallel: bool = False, nopython: bool = True, nogil: bool = False,
        cache: bool = False)\
        -> Callable[[Callable[..., Out]], Callable[..., Out]]:
    """This are the types of the decorator jit.
    Decorators with arguments return function, that takes a function as an
    argument. E.g. they return a decorator without parameters."""

    def decorator(func: Callable[..., Out]) -> Callable[..., Out]:
        """This is the unparametrized decorator"""

        @wraps(func)
        def wrapped_fun(*args: Any, **kwargs: Any) -> Out: ...

        return wrapped_fun

    return decorator


def prange(i: int) -> Iterable[int]: ...


@contextmanager
def objmode(**kwargs: Optional[str]) -> Iterator[None]: ...



#!/usr/bin/python
"""this set of unit-tests test the creation of plots by
'python-module'"""
from os import remove
from os.path import exists
from pathlib import Path

import numpy as np
from numpy import linspace
from pytest import mark

from scientific_plots.types_ import Vector
from scientific_plots.plot_settings import read_data_plot
from scientific_plots.default_plots import (
    plot, two_plots, two_axis_plots, plot_surface)


@mark.use_style
def test_default_plot(tmp_path: Path) -> None:
    """
    test the default plot creation by 'plot'
    """
    x_test: Vector = linspace(1., 100., 100)
    y_test: Vector = x_test**2

    x_label = "x-values"
    y_label = "y-values"
    filename = Path(".test_plot.tmp.pdf")
    filename2 = Path(".test_plot.tmp.png")
    plot(x_test, y_test, x_label, y_label, tmp_path / filename)
    assert exists(tmp_path / filename)

    # test data-plot creation
    assert exists(tmp_path / (filename.stem + ".csv"))
    data = read_data_plot(tmp_path / (filename.stem + ".csv"))
    assert data
    assert (list(data.values())[0][0] == x_test).all()
    assert (list(data.values())[0][1] == y_test).all()

    # remove plot to test another
    remove(tmp_path / filename)

    # logarithmic plot
    plot(x_test, y_test, x_label, y_label, tmp_path / filename, logscale=True)
    assert exists(tmp_path / filename)
    subfolder = ["sans_serif", "grayscale", "journal", "presentation"]
    for folder in subfolder:
        assert (
            exists(tmp_path / folder / filename)
            or exists(tmp_path / folder / filename2))


@mark.use_style
def test_default_plot_nan(tmp_path: Path) -> None:
    """
    Test the default plot creation by 'plot', if there are data points with
    nan.
    """
    x_test: Vector = linspace(2., 100., 50)
    y_test: Vector = x_test**2

    y_test[4] = np.nan
    y_test[6] = np.nan

    x_label = "x-values-nan"
    y_label = "y-values-nan"
    filename = Path(".test_plot_nan.tmp.pdf")
    plot(x_test, y_test, x_label, y_label, tmp_path / filename)
    assert exists(tmp_path / filename)


@mark.use_style
def test_two_plots(tmp_path: Path) -> None:
    """Test the creation of a twin-plot."""
    x_test: Vector = linspace(1., 100., 100)
    y_test1: Vector = x_test**2
    y_test2: Vector = x_test**.5
    filename = ".test_plot.tmp.pdf"
    two_plots(x_test, y_test1, "plot1",
              x_test, y_test2, "plot2",
              "x", "y",
              tmp_path / filename)
    assert exists(tmp_path / filename)


@mark.use_style
def test_twinx_plots(tmp_path: Path) -> None:
    """Test the creation of a twin-plot."""
    x_test: Vector = linspace(1., 100., 100)
    y_test1: Vector = x_test**2
    y_test2: Vector = x_test**.5
    filename = ".test_twin_plot.tmp.pdf"
    two_axis_plots(
        x_test, y_test1, "plot1",
        x_test, y_test2, "plot2",
        "x", "y", "y2",
        tmp_path / filename)
    assert exists(tmp_path / filename)

    remove(tmp_path / filename)
    # tick-plot
    two_axis_plots(
        x_test, y_test1, "plot1",
        x_test, y_test2, "plot2",
        "x", "y", "y2",
        tmp_path / filename,
        ticks=([1., 10.], ["a", "b"]))
    assert exists(tmp_path / filename)


@mark.use_style
def test_two_d_plot(tmp_path: Path) -> None:
    """Test the creation of a two dimensional surface plot."""
    x_test: Vector = linspace(1, 1e8, 1000)
    y_test: Vector = linspace(1, 1e8, 1000)
    x_test_grid, y_test_grid = np.meshgrid(
        x_test, y_test)
    z_grid = x_test_grid**2 + y_test_grid
    filename = Path(".test_two_d_plot.tmp.pdf")
    plot_surface(
        x_test_grid, y_test_grid, z_grid,
        "x-label", "y-label", "z-label",
        tmp_path / filename,
        log_scale=True)
    assert exists(tmp_path / filename)



"""Stub-Files for the Hurst-Module."""
from typing import Any, Tuple, Optional

from scientific_plots.types_ import Vector


def compute_Hc(x: Vector, kind: Optional[str] = "change",
               simplified: bool = True)\
    -> Tuple[float, Any, Any]: ...



"""Stub-file for scipy.interpolate. It also contains the typings of the used
functions."""
from __future__ import annotations

from typing import Callable, Optional, TypeVar, Union, overload, Literal

from scientific_plots.types_ import Matrix, Vector, Tensor

Input = TypeVar("Input", float, Matrix, Vector, Tensor)


def interp1d(x: Vector, y: Vector, fill_value: Optional[str] = "")\
    -> Callable[[Input], Input]: ...


class UnivariateSpline():
    """This class contains the spline-object of scipy."""

    def __init__(self, x: Vector, y: Vector,
                 k: int = 3, s: Optional[int] = None,
                 ext: Union[Literal[0], Literal[1], Literal[2], Literal[3],
                            Literal["extrapolate"], Literal["zeros"],
                            Literal["raise"], Literal["const"], str]
                 = "extrapolate",
                 check_finite: bool = False) -> None: ...

    def __call__(self, x: Input) -> Input: ...

    def derivative(self) -> UnivariateSpline: ...

    def roots(self) -> Vector: ...


class RectBivariateSpline:
    """This class contains a 2D spline of a surface."""

    def __init__(
        self,
        x: Union[Vector, list[float]],
        y: Union[Vector, list[float]],
        z: Union[Matrix, list[list[float]], list[Vector]],
        bbox: Optional[tuple[float, float, float, float]] = None,
        kx: int = 3,
        ky: int = 3,
        s: Optional[float] = 0) -> None: ...

    @overload
    def __call__(self,
                 x: Input,
                 y: Input,
                 dx: int = 0, dy: int = 0, *,
                 grid: Literal[False]) -> Input: ...

    @overload
    def __call__(self,
                 x: Union[Vector, list[float]],
                 y: Union[Vector, list[float]],
                 dx: int = 0, dy: int = 0, *,
                 grid: Literal[True]) -> Matrix: ...

    @overload
    def __call__(self,
                 x: Union[Vector, list[float]],
                 y: Union[Vector, list[float]],
                 dx: int = 0, dy: int = 0) -> Matrix: ...

    def integral(self,
                 xa: float, xb: float,
                 ya: float, yb: float) -> float: ...

    def ev(
        self, xi: Input, yi: Input, dx: Optional[int] = None,
        dy: Optional[int] = None) -> Input: ...

    def get_coeffs(self) -> list[float]: ...


class CubicSpline:
    """Types for the cubic spline function in scipy."""
    def __init__(
        self,
        x: Union[Vector, list[float]],
        y: Union[Vector, list[float]],
        axis: Optional[int] = 0,
        bc_type: Optional[Union[str, tuple[float, float]]] = None,
        extrapolate: Optional[Union[str, bool]] = None) -> None: ...

    def __call__(self, x: Input, nu: int = 0,
                 extrapolate: Optional[Union[bool, str]] = None) -> Input: ...

    def derivative(self) -> CubicSpline: ...

    def integrate(self,
                  a: float, b: float,
                  extrapolate: Optional[Union[str, bool]] = None) -> float: ...



#!/usr/bin/env python3
"""This is the stub-file for the window-functions in scipy. They can be used to
eliminate artifacts in the calculated Fourier-spectra."""
from scientific_plots.types_ import Vector


def kaiser(M: int, beta: float, sym: bool = True) -> Vector: ...



#!/usr/bin/env python3
"""
Stubs for the Python-Translator python module.
"""
from __future__ import print_function


class Translator():
    """
    Translator API.
    """

    def translate(self, input_str: str, lang1: str, lang2: str) -> str: ...



#!/usr/bin/env python3
"""
Test the design of the created plots by eye and by hand. For this, the plots
are placed in a persistent directory.
"""
from pathlib import Path

import matplotlib.pyplot as plt
from pytest import mark
from numpy import linspace, meshgrid
from scientific_plots.default_plots import (
    two_plots, three_axis_plots, plot, plot_surface)
from scientific_plots.plot_settings import apply_styles
from scientific_plots.types_ import Vector


LOCATION = Path("tests/plots/test_plot.pdf")
LOCATION2 = Path("tests/plots/test_plot2.pdf")
LOCATION3 = Path("tests/plots/test_plot3.pdf")
LOCATION4 = Path("tests/plots/test_plot_surface.pdf")
LOCATION5 = Path("tests/plots/test_plot_fallback.pdf")


@mark.use_style
def test_two_plots() -> None:
    """Test the creation of a twin-plot."""
    x_test: Vector = linspace(1., 100., 100)
    y_test1: Vector = x_test**2
    y_test2: Vector = x_test**.5
    filename = LOCATION
    filename.parent.mkdir(exist_ok=True)
    two_plots(x_test, y_test1, "plot1",
              x_test, y_test2, "plot2",
              "x", "y",
              filename)
    assert filename.exists()


@mark.use_style
def test_three_plots() -> None:
    """Test the creation of a plot with three y axis."""
    x_test: Vector = linspace(1., 100., 100)
    y_test1: Vector = x_test**2
    y_test2: Vector = x_test**.5
    y_test3: Vector = x_test**3 + 1
    filename = LOCATION2
    filename.parent.mkdir(exist_ok=True)
    three_axis_plots(
        x_test, y_test1, "1.",
        x_test, y_test2, "2.",
        x_test, y_test3, "3.",
        "x", "y1", "y2", "y3",
        filename, color=(3, 4, 5))
    assert filename.exists()


@mark.use_style
def test_skip_color() -> None:
    """Test the skipping of a color in the cycle."""
    x_test: Vector = linspace(1., 2., 100)
    y_test1: Vector = x_test**2
    filename = LOCATION3
    filename.parent.mkdir(exist_ok=True)
    plot(x_test, y_test1, "$x_2$", r"$\sqrt{y_c}$", filename, cycler=1)
    assert filename.exists()


@mark.use_style
def test_two_d_plot() -> None:
    """Test the creation of a two dimensional surface plot."""
    x_test: Vector = linspace(-10, 10, 100)
    y_test: Vector = linspace(-10, 10, 100)
    x_test_grid, y_test_grid = meshgrid(
        x_test, y_test)
    z_grid = x_test_grid**2 - y_test_grid
    filename = LOCATION4
    filename.parent.mkdir(exist_ok=True)
    plot_surface(
        x_test_grid, y_test_grid, z_grid,
        "x-label", "y-label", "z-label",
        LOCATION4)
    assert filename.exists()


@mark.use_style
def test_fallback() -> None:
    """Test plot creation using the fallback-style."""
    filename = LOCATION4
    filename.parent.mkdir(exist_ok=True)

    x_test: Vector = linspace(-10, 10, 100)
    y_test = x_test**2

    @apply_styles(_fallback=True)
    def simple_plot() -> None:
        """Create a simple plot."""
        plt.plot(x_test, y_test, label="test")
        plt.xlabel("$x$ [m]")
        plt.ylabel(r"$\sqrt{y_c}$ [1/m]")
        plt.legend()
        plt.tight_layout()
        plt.savefig(LOCATION5)

    simple_plot()

    assert filename.exists()



#!/usr/bin/env python3
"""
Type hints for the cyler class, which is included in matplotlib and used for
property cycles.
"""
from __future__ import annotations

from typing import Union, Any


class cycler:
    """Property cycle."""
    
    def __init__(
        self,
        color: Union[None,
                     list[tuple[float, float, float]],
                     list[str]] = None,
        linestyle: Union[None,
                         list[Union[str, tuple[int, tuple[int, ...]]]],
                         tuple[int, tuple[int, ...]]] = None): ...

    def __add__(self, other: cycler) -> cycler: ...

    def __mul__(self, other: cycler) -> cycler: ...

    def by_key(self) -> dict[str, Any]: ...



#!/usr/bin/python
"""This file contains the unit-tests for utilities in python_module"""
from os import remove
from pathlib import Path

from scientific_plots.utilities import write_file, read_file, running_average


def test_average() -> None:
    """test the running average"""
    x_test = [1. for _ in range(100)]
    for i in range(1, 10):
        assert running_average(x_test, i) == x_test


def test_read_write(tmp_path: Path) -> None:
    """test the read and the write functions of the utilities"""
    x_list = list(float(x) for x in range(20))
    y_list = [x**2 for x in x_list]
    filename = ".test.tmp"
    write_file(tmp_path / filename, x_list, y_list)
    x_new, y_new = read_file(tmp_path / filename)
    remove(tmp_path / filename)
    assert x_list == x_new
    assert y_list == y_new



#!/usr/bin/env python3
"""
Stubs for matplotlib-ticker.
"""
from __future__ import annotations

from typing import Callable, Any


def StrMethodFormatter(format_str: str) -> str: ...


def FuncFormatter(func: Callable[[float, Any], str]) -> None: ...


class Locator:
    """Base class for locators."""


class MaxNLocator(Locator):
    """N ticks on an axis."""
    nbins: int

    def __init__(self, nbins: int): ...


class AutoLocator(Locator):
    """Autolocator."""
    nbins: int

    def __init__(self, nbins: int): ...



#!/usr/bin/env python
"""
This test-file contains test for the behaviour of new types, which are defined
in the python module. Numpy-arrays often create errors, when they are used by
static type checkers, because the multiplication sometimes returns a
NoReturn-type.
"""
import numpy as np

from scientific_plots.types_ import Matrix


def test_static_checking() -> None:
    """Test matrix-multiplications."""
    N = 20
    M = 31
    x: Matrix = np.random.rand(N, M)
    z = x**2 + x*x  # type: ignore
    # the type-checkings needs to be ignored, because we cannot use
    # numpy-version 1.21, where proper type-checking is introduced. This is due
    # to a dependency of numba.
    assert z is not None



#!/usr/bin/env python
"""
Stub-file for usage with matplot-lib. This file registers color-maps for later
usage.
"""
from typing import Union

from .colors import LinearSegmentedColormap, ListedColormap


def register_cmap(name: str, cmap: Union[LinearSegmentedColormap,
                  ListedColormap]) -> None: ...


class ColormapRegistry:
    """Container for colormaps."""
    def register(
        self,
        cmap: Union[LinearSegmentedColormap, ListedColormap])\
        -> None: ...



#!/usr/bin/env python
"""
These are the stub-files for matplot-lib. This stub-module also contains the
sub-modules, which are sometimes needed when using matplot-lib functions. This
stub-file provides the types, so mypy can use static-type-checking on scripts
or modules, which use matplot-lib. No function is implemented in this folder.
"""
from __future__ import annotations
from typing import Union, overload, Literal

from cycler import cycler

from .cm import ColormapRegistry


def use(backend: str) -> None: ...


def rc(object_: str, **kwargs: Union[
        str,
        bool,
        dict[str, int],
        int]) -> None: ...


class RCParams:
    """Default parameters of the creation of plots in matplot-lib."""

    def update(self, params: Union[dict[str,
                                        Union[int, list[float], str,
                                              tuple[float, float]]
                                        ],
                                   RCParams]) -> None: ...

    def __setitem__(self, key: str, value: Union[list[str],
                                                 int,
                                                 str,
                                                 tuple[float, float],
                                                 cycler])\
        -> None: ...

    @overload
    def __getitem__(self, key: Literal["axes.prop_cycle"]) -> cycler: ...

    @overload
    def __getitem__(self, key: str) -> Union[list[str], cycler]: ...


rcParams: RCParams


rcParamsDefault: RCParams


colormaps: ColormapRegistry



#!/usr/bin/env python3
"""This is a python-module for scientific plotting.
It contains functions and methods, which are
useful across several different projects and repositories. It also contains
stub-files with several data-type annotations for scipy and matplot-lib
functions."""

__version__ = "1.7.2"



#!/usr/bin/env python3
"""this module contains functions, which are useful for
analysing data """
from __future__ import annotations
from math import sqrt
from typing import Union, overload

from scipy.stats import linregress
import numpy as np

from .types_ import Vector, Matrix


ArrayLike = Union[list[float], Vector]
MatrixLike = Union[list[list[float]], Matrix]


@overload
def get_hrms(X: ArrayLike, Y: ArrayLike) -> float: ...


@overload
def get_hrms(X: ArrayLike, Y: MatrixLike) -> Vector: ...


def get_hrms(X: ArrayLike, Y: Union[ArrayLike, MatrixLike])\
        -> Union[float, Vector]:
    """calculate R_q or h_rms based on real space data
    calculate a single R_q if Y is a plain
    array and take the average if it is an array of arrays"""
    if isinstance(Y, list):
        Y = np.array(Y)
    if isinstance(X, list):
        X = np.array(X)
    if Y.ndim == 1:
        Y = np.array([Y])
    h_rms = []
    for y in Y:
        res = linregress(X, y)
        alpha = res[0]
        beta = res[0]
        y_lin = [y_i - alpha * x - beta for x, y_i in zip(X, y)]
        mean_y = sum(y_lin) / len(y_lin)
        y_lin = [y_i - mean_y for y_i in y_lin]
        h_rms += [sqrt(sum(y_lin_i ** 2 for y_lin_i in y_lin))]
    mean_h_rms = sum(h_rms) / len(h_rms)
    return mean_h_rms



#!/usr/bin/env python
"""
Stubs for scipy.stats. Here are type-annotations for scipy's
linear-regression-like methods.
"""
from typing import Tuple, Union

from scientific_plots.types_ import Vector


Input = Union[Vector, list[float]]


def linregress(x: Input, y: Input) -> Tuple[float, float]: ...



#!/usr/env python
# pylint: disable=unsubscriptable-object
"""This module contains the types, which can be used for static
type-checking in Python_Module."""
from __future__ import annotations
from typing import TYPE_CHECKING, Tuple, Any

import numpy as np
from pytest import FixtureRequest

if TYPE_CHECKING:

    NTensor = np.ndarray[Tuple[int, int, int, int], np.dtype[np.float64]]
    Tensor = np.ndarray[Tuple[int, int, int], np.dtype[np.float64]]
    Matrix = np.ndarray[Tuple[int, int], np.dtype[np.float64]]
    Vector = np.ndarray[Tuple[int], np.dtype[np.float64]]
else:
    NTensor = np.ndarray
    Tensor = np.ndarray
    Matrix = np.ndarray
    Vector = np.ndarray


class Request(FixtureRequest):
    """This is the pytest-fixture request plus the missing attribute 'param'.
    It contains the given parameters to this object and thus this fixture."""

    param: Any



#!/usr/bin/env python
"""
This are the stubs for the submodule "style" of matplotlib.pyplot.
"""
from contextlib import contextmanager
from collections.abc import Generator


def use(style: str = "default") -> None: ...


@contextmanager
def context(style: str) -> Generator[None, None, None]: ...



#!/usr/bin/env python
"""
Stub-file for scipy.optimize. It contains only the typing-annotation for this
module.
"""

from typing import Tuple, Callable, List, Optional, TypeVar, Union, Any

from scientific_plots.types_ import Matrix, Vector, Tensor

Input = TypeVar("Input", float, list[float], Matrix, Vector, Tensor)


def curve_fit(
    func: Callable[..., Input],
    x: Union[Vector, list[float]],
    y: Union[Vector, list[float]],
    p0: Optional[Union[List[float], tuple[float]]] = None,
    bounds: Optional[Any] = None)\
    -> Tuple[Vector, Matrix]: ...


def brentq(
    func: Callable[[Input], Input],
    start: float,
    end: float,
    maxiter: int = 1000,
    xtol: Optional[float] = None,
    rtol: Optional[float] = None) -> float: ...


class OptimizeResult:
    """The class contains the result of the root-finding algorithm."""
    x: Vector
    success: bool
    status: int
    message: str


def root(
    func: Callable[[Input], Input],
    start: float) -> OptimizeResult: ...



"""This is a stub-file for scipy.special. It contains the typing for needed
function and allows the type-checker to work."""

from typing import TypeVar

from scientific_plots.types_ import Matrix, Vector, Tensor


Input = TypeVar("Input", float, Matrix, Vector, Tensor)


def expi(x: Input) -> Input: ...


def i0(x: Input) -> Input: ...


def erf(x: Input) -> Input: ...



#!/usr/bin/env python
"""This is the stub for the figure-type in matplot-lib."""
from __future__ import annotations

from pathlib import Path
from typing import Optional, Tuple, Iterable, Any, Union

from .pyplot import Axes


class Canvas:
    """Canvas, which contains the figure."""

    def draw(self) -> None: ...

    def fush_events(self) -> None: ...


class Figure:

    axes: Iterable[Axes]

    canvas: Canvas

    def __init__(
        self, figsize: Optional[Tuple[float, float]] = None) -> None: ...

    def gca(self) -> Axes: ...

    def set_size_inches(self, x: float, y: float) -> None: ...

    def subplots_adjust(self, bottom: Optional[float] = None,
                        right: Optional[float] = None,
                        left: Optional[float] = None,
                        top: Optional[float] = None,
                        ) -> None: ...

    def add_subplot(self, position: int = 111,
                    projection: str = "2d") -> Axes: ...

    def set_label(self, label: str) -> None: ...

    def text(self, x: float, y: float, s: str) -> None: ...

    def legend(self, *args: Any, **kwargs: Any) -> None: ...

    def savefig(
        self, fname: Union[str, Path], *,
        dpi: Union[str, float, int, None] = "figure",
        bbox_inches: Optional[Union[tuple[float, float], str]] = None,
        pad_inches: float = 0.1, backend: Optional[str] = None
        ) -> None: ...

    def tight_layout(self) -> None: ...



"""This module contains the typing-stubs for the fft-module of scipy. They are
needed for type-checking by mypy."""
from typing import Tuple, TypeVar

from scientific_plots.types_ import Matrix, Vector


def fft(x: Vector) -> Vector: ...


def rfft(x: Vector) -> Vector: ...


def fft2(x: Matrix, s: Tuple[int, ...]) -> Matrix: ...


def rfft2(x: Matrix, s: Tuple[int, ...]) -> Matrix: ...


SmallDimension = TypeVar("SmallDimension", Matrix, Vector)


def fftshift(x: SmallDimension) -> SmallDimension: ...


def fftfreq(n: int, d: float = 1.0) -> Vector: ...


def rfftfreq(n: int, d: float = 1.0) -> Vector: ...



#!/usr/bin/env python
"""
Yet another stub-file for matplot-lib.
"""
from typing import List, Tuple, Dict, Optional


class ColorMap:
    """BaseClass."""


class ListedColormap(ColorMap):
    def __init__(self, color_list: List[Tuple[int, int, int]],
                 name: Optional[str] = None) -> None: ...


class LinearSegmentedColormap(ColorMap):
    def __init__(self, name: str,
                 color_dict: Dict[str,
                                  List[Tuple[float,
                                             Optional[float],
                                             Optional[float]
                                             ]
                                       ]
                                  ]
                 ) -> None: ...



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""
This module contains usefule utilities for many
different python scripts
"""
from __future__ import print_function

import re
from functools import wraps
from subprocess import Popen, PIPE
from sys import __stdout__
from os import mkdir
from os.path import dirname, exists
from typing import Iterable, Optional, List, Callable, TypeVar, Union, Any
from pathlib import Path
from collections import OrderedDict

from urllib3.exceptions import MaxRetryError
from python_translator import Translator


def running_average(X: List[float], n: int) -> List[float]:
    """creates a running average of X with n entries in both dircetions"""
    X_new = []
    for i in range(n):
        X_new += [sum(X[0: i + 1]) / (i + 1)]
    for i in range(n, len(X) - n):
        X_new += [sum(X[i - n: i + n + 1]) / (2 * n + 1)]
    for i in range(len(X) - n, len(X)):
        X_new += [sum(X[2 * i - len(X) + 1:]) / ((len(X) - (i + 1)) * 2 + 1)]
    return X_new


def quality_of_fit(X: List[float], Y: List[float]) -> float:
    """calculates the quality of a fit 'bestimmtheitsmass'"""
    mean_x = sum(X) / len(X)
    return sum(x ** 2 - mean_x ** 2 for x in X)\
        / sum(x ** 2 - mean_x ** 2 for x in Y)


Return = TypeVar("Return")


def print_redirect(f: Callable[..., Return]) -> Callable[..., Return]:
    """
    wraps print to print to both stdout and __stdout__
    the reason for doing that is that abaqus replaces
    stdout and prints everything to abaqus
    this way everything can be seen both in the command line and in abaqus
    """

    @wraps(f)
    def inner_function(*args: Any, **argv: Any) -> Return:
        f(*args, **argv)
        for i, arg in enumerate(args):
            if isinstance(type, arg):
                del arg[i]
        for _, val in argv.items():
            if isinstance(type, val):
                del val
        argv["file"] = __stdout__
        f(*args, **argv)
        try:
            return f(*args, **argv)
        except TypeError:
            print("The argument 'file' is twice in a print statement")
            raise

    return inner_function


def filter_function(x: str) -> bool:
    """filter function in order to remove unused
    arguments from this script before parsing"""
    if ":" in x or "rwth" in x:
        return False
    analysed_item = x.replace("--", ".")
    analysed_item = analysed_item.replace("-", "!")
    if len(analysed_item) > 2\
            and "." not in analysed_item\
            and "!" in analysed_item:
        return False
    return True


def filter_argv(X: list[str]) -> list[str]:
    """removes unessecery entries from
    argv which have been generated by abaqus"""
    Y: list[str] = list(X)
    for i, x in enumerate(X):
        if not filter_function(x):
            del Y[i]
            if i < len(Y):
                del Y[i + 1]
    return Y


def dumb_plot(X: list[float], Y: list[float],
              title: str = "title", log: bool = False) -> None:
    """plot X and Y using dumb gnuplot"""
    try:
        with Popen(["gnuplot"], stdin=PIPE) as gnuplot:
            assert gnuplot, """Gnuplot could not be started."""
            assert gnuplot.stdin, """Input to Gnuplot could not be found."""
            gnuplot.stdin.write(bytes("set term dumb 79 25\n", "utf-8"))
            if log:
                gnuplot.stdin.write(bytes("set logscale xy\n", "utf-8"))
            gnuplot.stdin.write(
                bytes(
                    f"""plot '-' using 1:2 title '{title}'
                    with linespoints \n""",
                    "utf-8"
                )
            )
            for x, y in zip(X, Y):
                gnuplot.stdin.write(bytes(f"{x} {y}\n", "utf-8"))
            gnuplot.stdin.write(bytes("e\n", "utf-8"))
            gnuplot.stdin.flush()
    except FileNotFoundError:
        print("There is no installed instance of gnuplot")
        return


def read_file(filename: Union[str, Path],
              header: int = 0,
              type_: type = float) -> tuple[list[float], ...]:
    """read a file with given filename into
    a python-tuple. Skip n='header' lines in the beginning
    of the document"""
    # reads a given file and returns lists
    lists: list[list[float]] = []
    with open(filename, "r", encoding="utf-8") as input_file:
        for i, line in enumerate(input_file):
            if i < header:
                continue
            lists += [[type_(v) for v in line.split()]]
    return tuple(list(v) for v in zip(*lists))


def write_file(filename: Union[Path, str],
               *args: Iterable[float],
               title: Optional[str] = None) -> None:
    """write the rows given in 'arga' to a file"""
    # write data to a file
    if not any(args[0]):
        raise Exception("Tried to write an empty row to a file")
    if isinstance(filename, str):
        filename = Path(filename)
    with open(filename, "w", encoding="utf-8") as output_file:
        if title is not None:
            output_file.write(title + "\n")
        for row in zip(*args):
            output_file.write(" ".join(str(r) for r in row) + "\n")


def mkdir_p(foldername: str) -> None:
    """creates a new folder if the folder does not exist"""
    try:
        if exists(foldername):
            return
        mkdir(foldername)
    except IOError as e:
        # recursive creation of needed folders
        if e.errno != 2:
            raise
        mkdir_p(dirname(foldername))
        mkdir_p(foldername)


def trash_remover(func: Callable[..., tuple[Return, ...]])\
        -> Callable[..., Return]:
    """only keeps the first output of a given function"""

    @wraps(func)
    def inner_function(*args: Any, **kwargs: Any) -> Return:
        result = func(*args, **kwargs)
        return result[0]

    return inner_function


def use_translator(string: str, lang1: str, lang2: str) -> str:
    """
    Translate the given input string from lang2 to lang1 using google
    translate.
    """
    # one letter strings
    if len(string) <= 1:
        return string

    if r"\times" in string:
        return string

    try:
        translator = Translator()
        result: str
        result = translator.translate(string, lang1, lang2)
    except (OSError, MaxRetryError):
        return string
    return result


def translate(string: str, reverse: bool = False) -> str:
    """translates a string from english to german
    @input string any string wich contains specific english words
    @input reverse translate ger->en if set to true
    @return the german translation of the same string"""

    if reverse:
        string = re.sub(r'(\d+),(\d+)', r'\1\.\2', string)
    else:
        string = re.sub(r'(\d+)\.(\d+)', r'\1,\2', string)

    _dict: dict[str, str] = OrderedDict({
        "leakage": "Leckage",
        "contact pressure": "Kontaktdruck",
        "fluid pressure": "Fluiddruck",
        "pressure": "Druck",
        "density": "Dichte",
        "roundness": "Rundheit",
        "eccentricity": r"Exzentrizit\"at",
        "contact area": r"Kontaktoberfl\"ache",
        "area": r"Fl\"ache",
        "maximal": "Maximale",
        "time": "Zeit",
        "normal-force": "Normalkraft",
        "normal force": "Normalkraft",
        "total force": "Gesamtkraft",
        "force": "Kraft",
        "distance": "Abstand",
        "position": "Position",
        "contact broadness": "Kontaktbreite",
        "broadness": "Breite",
        "contact": "Kontakt-",
        "seat": "Sitz",
        "ball": "Kugel",
        "high": "hoch",
        "low": "tief",
        "elastic": "elastisch",
        "plastic": "plastisch",
        "angle": "Winkel",
        "degree": "Grad",
        "deg": "Grad",
        "hard": "hart",
        "soft": "weich",
        "fit": "Fit",
        "data": "Messwerte",
        "velocity": "Geschwindigkeit",
        "measurement": "Messung",
        "experiment": "Experiment",
        "simulation": "Simulation",
        "analytical": "analytisch",
        "signal": "Signal",
        "valve control": "Ventilansteuerung",
        "off": "zu",
        "on": "auf",
        "status": "Zustand",
        "valve": "Ventil",
        "relative pressure": "Relativdruck",
        "absolute pressure": "Absolutdruck",
        "relative": "relativ",
        "absolute": "absolut",
        "plot": "Graph"
    })
    if not reverse and r"\times" not in string:
        for key, value in _dict.items():
            if key in string.lower():
                string = re.sub(key, value, string, flags=re.IGNORECASE)
                break
        else:
            string = use_translator(string, "german", "english")
    else:
        for key, value in _dict.items():
            if value.lower() in string.lower():
                string = string.lower().replace(value.lower(), key.lower())
                break
        else:
            string = use_translator(string, "english", "german")
    return string



#!/usr/bin/env python
"""
Stub-files for scipy.signal. This module contains the type-annotations for
scipy's library for fourier-transformations and signal analysis.
"""
from typing import Tuple, Optional, Union, overload

from scientific_plots.types_ import Vector


def welch(
    Y: Union[Vector, list[float]],
    fs: Optional[float] = None,
    scaling: str = "density",
    window: Optional[str] = "hamming",
    nperseg: int = 10,
    detrend: bool = False) -> Tuple[Vector, Vector]: ...


def periodogram(
    Y: Union[Vector, list[float]],
    fs: Optional[float] = None,
    scaling: str = "density",
    window: Optional[str] = "hamming",
    detrend: bool = False) -> Tuple[Vector, Vector]: ...


Single = Union[Vector, list[float]]
Double = Union[tuple[Vector, Vector], tuple[list[float], list[float]]]


@overload
def savgol_filter(
    values: Single, window: int, order: int) -> Vector: ...


@overload
def savgol_filter(
    values: Double, window: int, order: int)\
        -> tuple[Vector, Vector]: ...



#!/usr/bin/env Python3
# pylint: disable=too-many-locals,too-many-arguments
"""
This module contains a few functions, which can be used to
generate plots quickly and with useful defaults"""
from __future__ import annotations
from pathlib import Path
from functools import wraps
from typing import TypeVar, List, Tuple, Union, Callable, Optional
from warnings import warn, filterwarnings, catch_warnings
from textwrap import dedent

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import numpy as np
from numpy import amin, amax

from .plot_settings import apply_styles, rwth_cycle
from .types_ import Vector, Matrix

mpl.use("Agg")

In = TypeVar("In", List[float], Tuple[float],
             Vector)

In2D = TypeVar("In2D", list[list[float]], list[Vector], tuple[Vector],
               Matrix)


def fix_inputs(input_1: In, input_2: In)\
        -> tuple[Vector, Vector]:
    """
    Remove nans and infinities from the input vectors.

    Parameters
    ---------
    input_1, input_2:
        X/Y-axis data of plot

    Returns
    ------
    New vectors x and y with nans removed.
    """
    if len(input_1) != len(input_2):
        raise ValueError(
            "The sizes of the input vectors are not the same.")
    nan_count = np.count_nonzero(np.isnan(input_2))
    inf_count = np.count_nonzero(np.isinf(input_2))
    if nan_count != 0 or inf_count != 0:
        new_input_1 = np.empty(len(input_1)-nan_count-inf_count)
        new_input_2 = np.empty(len(new_input_1))
        position = 0
        for x_input, y_input in zip(input_1, input_2):
            if not np.isnan(y_input) or np.isinf(y_input):
                new_input_1[position] = x_input
                new_input_2[position] = y_input
                position += 1
        return new_input_1, new_input_2

    return np.array(input_1), np.array(input_2)


def check_inputs(input_1: In, input_2: In, label_1: str, label_2: str)\
        -> bool:
    """
    Check the input vectors to see, if they are large enough.

    Parameters
    ---------
    input_1, input_2:
        X/Y-axis data of plot

    label_1, label_2:
        Labels of the X/Y axis

    Returns
    ------
    True, if the plot can be created.
    """
    if len(input_1) <= 1 or len(input_2) <= 1:
        warn(
            "There are not enough points in the following plots:"
            f"label1: {label_1} label2: {label_2}. It cannot be drawn.")
        return False

    if min(input_1) == max(input_1):
        warn(
            "The area of the x-axis is not large enough in the following plot:"
            f"label1: {label_1} label2: {label_2}. It cannot be drawn.")
        return False

    if min(input_2) == max(input_2):
        warn(
            "The area of the y-axis is not large enough in the following plot:"
            f"label1: {label_1} label2: {label_2}. It cannot be drawn.")
        return False

    infinity = np.isinf(input_1).any() or np.isinf(input_2).any()
    if infinity:
        warn(dedent(f"""There are infinities in the data of the following plot:
             label1: {label_1}, label2: {label_2}. It cannot be drawn."""),
             RuntimeWarning)
        return False

    nan = np.isnan(input_1).any() or np.isnan(input_2).any()
    if nan:
        warn(dedent(f"""There are nans in the data of the following plot:
             label1: {label_1}, label2: {label_2}. It cannot be drawn."""),
             RuntimeWarning)
        return False

    return True


@apply_styles
def plot_fit(X: In, Y: In,
             fit_function: Callable[..., float],
             xlabel: str, ylabel: str, filename: Union[str, Path], *,
             args: Optional[Tuple[float]] = None,
             logscale: bool = False) -> None:
    """Creates a plot of data and a fit and saves it to 'filename'."""
    X, Y = fix_inputs(X, Y)  # type: ignore
    if not check_inputs(
            X, Y, xlabel, ylabel):
        return

    n_fit = 1000

    _fit_function: Callable[[float], float]
    if args is not None:

        @wraps(fit_function)
        def _fit_function(x: float) -> float:
            """This is the function, which has been fitted"""
            return _fit_function(x, *args)
    else:
        _fit_function = fit_function

    plt.plot(X, Y, label="data")
    X_fit = [min(X) + (max(X) - min(X)) * i / (n_fit - 1)
             for i in range(n_fit)]
    Y_fit = [_fit_function(x) for x in X_fit]
    plt.plot(X_fit, Y_fit, label="fit")
    if logscale:
        plt.xscale("log")
        plt.yscale("log")
    plt.xlim(min(X), max(X))
    if logscale:
        plt.ylim(min(Y) * 0.97, max(Y) * 1.02)
    else:
        plt.ylim(
            min(Y) - (max(Y) - min(Y)) * 0.02,
            max(Y) + (max(Y) - min(Y)) * 0.02
        )
    plt.ylabel(ylabel)
    plt.xlabel(xlabel)
    plt.legend()
    plt.tight_layout()
    plt.savefig(filename)
    plt.close()


@apply_styles(three_d=True)
def plot_surface(X: In2D, Y: In2D, Z: In2D,
                 xlabel: str, ylabel: str, zlabel: str,
                 filename: Union[str, Path], *,
                 log_scale: bool = False,
                 set_z_lim: bool = True,
                 colorscheme: str = "rwth_gradient",
                 figsize: tuple[float, float] = (4.33, 3.5),
                 labelpad: Optional[float] = None,
                 nbins: Optional[int] = None) -> None:
    """create a 2D surface plot of meshgrid-like valued Xs, Ys and Zs"""
    if not check_inputs(
            np.array(X).flatten(),
            np.array(Z).flatten(), xlabel, zlabel):
        return
    fig = plt.figure()
    ax = fig.add_subplot(projection="3d")
    fig.subplots_adjust(left=-0.02, right=0.75, bottom=0.15, top=0.98)
    ax.plot_surface(X, Y, Z, cmap=colorscheme)
    ax.set_box_aspect(aspect=None, zoom=.8)

    if labelpad is None:
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_zlabel(zlabel, rotation=90)
    else:
        ax.set_xlabel(xlabel, labelpad=labelpad)
        ax.set_ylabel(ylabel, labelpad=labelpad)
        ax.set_zlabel(zlabel, rotation=90, labelpad=labelpad)

    assert ax.zaxis is not None

    ax.set_xlim(amin(X), amax(X))  # type: ignore
    ax.set_ylim(amin(Y), amax(Y))  # type: ignore

    if set_z_lim:
        if not log_scale:
            ax.set_zlim(
                amin(Z) - (amax(Z) - amin(Z)) * 0.02,  # type: ignore
                amax(Z) + (amax(Z) - amin(Z)) * 0.02  # type: ignore
            )
        else:
            ax.set_zlim(
                amin(Z) * 0.97, amax(Z) * 1.02)  # type: ignore

    if log_scale:
        ax.set_yscale("log")
        ax.set_xscale("log")
        ax.set_zscale("log")

    for spine in ax.spines.values():
        spine.set_visible(False)

    ax.xaxis.pane.set_alpha(0.3)
    ax.yaxis.pane.set_alpha(0.3)
    ax.zaxis.pane.set_alpha(0.3)

    if nbins is not None:
        ax.xaxis.set_major_locator(
            MaxNLocator(nbins)
        )
        ax.yaxis.set_major_locator(
            MaxNLocator(nbins)
        )

    fig.set_size_inches(*figsize)

    with catch_warnings():
        filterwarnings("ignore", message=".*Tight layout")
        plt.tight_layout()
        plt.savefig(filename)

    plt.close()


@apply_styles
def plot(X: In, Y: In, xlabel: str, ylabel: str,
         filename: Union[Path, str], *, logscale: bool = False,
         ylim: Optional[tuple[float, float]] = None,
         yticks: bool = True, cycler: int = 0) -> None:
    """Create a simple 1D plot"""
    X, Y = fix_inputs(X, Y)  # type: ignore
    if not check_inputs(
            X, Y, xlabel, ylabel):
        return
    if len(X) <= 1 or len(Y) <= 1:
        raise ValueError(
            f"The data for plot {filename} contains empty rows!")

    if cycler > 0:
        for _ in range(cycler):
            plt.plot([], [])
    plt.plot(X, Y, linestyle="-")
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    if logscale:
        plt.xscale("log")
        plt.yscale("log")
        if ylim is None:
            plt.ylim(min(Y) * 0.97, max(Y) * 1.02)
    elif ylim is None:
        plt.ylim(
            min(Y) - (max(Y) - min(Y)) * 0.02,
            max(Y) + (max(Y) - min(Y)) * 0.02
        )
    if ylim is not None:
        plt.ylim(*ylim)
    plt.xlim(min(X), max(X))
    if not yticks:
        plt.yticks([])
    plt.tight_layout()
    plt.savefig(filename)
    plt.close()


@apply_styles
def two_plots(x1: In, y1: In, label1: str,
              x2: In, y2: In, label2: str,
              xlabel: str, ylabel: str,
              filename: Union[Path, str], *,
              logscale: bool = False, cycle: int = 0,
              color: tuple[int, int] = (0, 1),
              outer: bool = False) -> None:
    """Create a simple 1D plot with two different graphs inside of a single
    plot and a single y-axis.

    Keyword arguments:
    cycle -- skip this many colours in the colour-wheel before plotting
    color -- use these indeces in the colour-wheel when creating a plot
    outer -- use the outer limits on the x-axis rather than the inner limit
    """
    x1, y1 = fix_inputs(x1, y1)  # type: ignore
    x2, y2 = fix_inputs(x2, y2)  # type: ignore

    if not (
            check_inputs(x1, y1, xlabel, label1)
            or check_inputs(x2, y2, xlabel, label2)):
        return
    if len(x1) <= 1 or len(y1) <= 1 or len(y2) <= 1 or len(x2) <= 1:
        raise ValueError(
            f"The data for plot {filename} contains empty rows!")

    if cycle > 0:
        color = (color[0] + cycle, color[1] + cycle)

    # access colour
    prop_cycle = plt.rcParams["axes.prop_cycle"]
    try:
        linestyle = prop_cycle.by_key()["linestyle"]
    except KeyError:
        linestyle = rwth_cycle.by_key()["linestyle"]
    colors = prop_cycle.by_key()["color"]

    if max(color) >= len(colors):
        colors += colors
        linestyle += linestyle

    plt.plot(x1, y1, label=label1,
             color=colors[color[0]],
             linestyle=linestyle[0])

    plt.plot(x2, y2, label=label2,
             color=colors[color[1]],
             linestyle=linestyle[1])
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    min_ = min(min(y1), min(y2))
    max_ = max(max(y1), max(y2))
    if not logscale:
        plt.ylim(
            min_ - (max_ - min_) * 0.02,
            max_ + (max_ - min_) * 0.02
        )
    else:
        plt.xscale("log")
        plt.yscale("log")
        plt.ylim(
            min_ * 0.97, max_ * 1.02)

    if outer:
        plt.xlim(min(min(x1), min(x2)),
                 max(max(x1), max(x2)))
    else:
        plt.xlim(max(min(x1), min(x2)),
                 min(max(x1), max(x2)))

    plt.legend()
    plt.tight_layout()
    plt.savefig(filename)
    plt.close()


@apply_styles
def three_plots(x1: In, y1: In, label1: str,
                x2: In, y2: In, label2: str,
                x3: In, y3: In, label3: str,
                xlabel: str, ylabel: str,
                filename: Union[Path, str], *,
                logscale: bool = False,
                xmin: Optional[float] = None,
                xmax: Optional[float] = None) -> None:
    """Create a simple 1D plot with three different graphs inside of a single
    plot and a single y-axis."""
    x1, y1 = fix_inputs(x1, y1)  # type: ignore
    x2, y2 = fix_inputs(x2, y2)  # type: ignore
    x3, y3 = fix_inputs(x3, y3)  # type: ignore
    if not (
            check_inputs(x1, y1, xlabel, label1)
            or check_inputs(x2, y3, xlabel, label1)
            or check_inputs(x3, y3, xlabel, label3)):
        return

    if any(len(x) <= 1 for x in (x1, x2, y1, y2, x3, y3)):
        raise ValueError(
            f"The data for plot {filename} contains empty rows!")

    plt.plot(x1, y1, label=label1)
    plt.plot(x2, y2, label=label2, linestyle="dashed")
    plt.plot(x3, y3, label=label3, linestyle="dotted")
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    min_ = min(min(y1), min(y2), min(y3))
    max_ = max(max(y1), max(y2), max(y3))
    if not logscale:
        plt.ylim(
            min_ - (max_ - min_) * 0.02,
            max_ + (max_ - min_) * 0.02
        )
    else:
        plt.xscale("log")
        plt.yscale("log")
        plt.ylim(
            min_ * 0.97, max_ * 1.02)
    if xmin is not None and xmax is not None:
        plt.xlim(xmin, xmax)
    else:
        plt.xlim(min(x1), max(x1))
    plt.legend()
    plt.tight_layout()
    plt.savefig(filename)
    plt.close()


@apply_styles
def two_axis_plots(x1: In, y1: In, label1: str,
                   x2: In, y2: In, label2: str,
                   xlabel: str, ylabel: str,
                   ylabel2: str,
                   filename: Union[Path, str], *,
                   ticks: Optional[tuple[list[float], list[str]]] = None,
                   xlim: Optional[tuple[float, float]] = None,
                   color: tuple[int, int] = (0, 1))\
        -> None:
    """Create a simple 1D plot with two different graphs inside of a single
    plot with two y-axis.
    The variable "ticks" sets costum y-ticks on the second y-axis. The first
    argument gives the position of the ticks and the second argument gives the
    values to be shown.
    Color selects the indeces of the chosen color-wheel, which should be taken
    for the different plots. The default is (1,2)."""
    x1, y1 = fix_inputs(x1, y1)  # type: ignore
    x2, y2 = fix_inputs(x2, y2)  # type: ignore
    if not check_inputs(
            y1, y2, label1, label2):
        return
    if len(x1) <= 1 or len(y1) <= 1 or len(y2) <= 1 or len(x2) <= 1:
        raise ValueError(
            f"The data for plot {filename} contains empty rows!")

    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    # access colour
    prop_cycle = plt.rcParams["axes.prop_cycle"]
    try:
        linestyle = prop_cycle.by_key()["linestyle"]
    except KeyError:
        linestyle = rwth_cycle.by_key()["linestyle"]
    colors = prop_cycle.by_key()["color"]

    if max(color) >= len(colors):
        colors += colors
        linestyle += linestyle

    # first plot
    lines = ax1.plot(x1, y1, label=label1,
                     color=colors[color[0]],
                     linestyle=linestyle[0])
    ax1.set_xlabel(xlabel)
    ax1.set_ylabel(ylabel)
    ax1.set_ylim(
        min(y1) - (max(y1) - min(y1)) * 0.02,
        max(y1) + (max(y1) - min(y1)) * 0.02
    )

    # second plot
    ax2 = ax1.twinx()
    lines += ax2.plot(x2, y2, label=label2,
                      color=colors[color[1]],
                      linestyle=linestyle[1])
    ax2.set_ylabel(ylabel2)
    ax2.set_ylim(
        min(y2) - (max(y2) - min(y2)) * 0.02,
        max(y2) + (max(y2) - min(y2)) * 0.02
    )

    # general settings
    if xlim is None:
        plt.xlim(min(x1), max(x1))
    else:
        plt.xlim(*xlim)
    labels = [line.get_label() for line in lines]
    plt.legend(lines, labels)
    # ticks
    if ticks is not None:
        ax2.set_yticks(ticks[0])
        ax2.set_yticklabels(ticks[1])
    plt.tight_layout()
    plt.savefig(filename)
    plt.close()


def make_invisible(ax: plt.Axes) -> None:
    """Make all patch spines invisible."""
    ax.set_frame_on(True)
    ax.patch.set_visible(False)
    for spine in ax.spines.values():
        spine.set_visible(False)


@apply_styles
def three_axis_plots(x1: In, y1: In, label1: str,
                     x2: In, y2: In, label2: str,
                     x3: In, y3: In, label3: str,
                     xlabel: str, ylabel: str,
                     ylabel2: str, ylabel3: str,
                     filename: Union[Path, str], *,
                     ticks: Optional[tuple[list[float], list[str]]] = None,
                     xlim: Optional[tuple[float, float]] = None,
                     color: tuple[int, int, int] = (0, 1, 2),
                     legend: bool = True)\
        -> None:
    """Create a simple 1D plot with two different graphs inside of a single
    plot with two y-axis.
    The variable "ticks" sets costum y-ticks on the second y-axis. The first
    argument gives the position of the ticks and the second argument gives the
    values to be shown.
    Color selects the indeces of the chosen color-wheel, which should be taken
    for the different plots. The default is (1,2)."""
    # pylint: disable=R0915
    x1, y1 = fix_inputs(x1, y1)  # type: ignore
    x2, y2 = fix_inputs(x2, y2)  # type: ignore
    x3, y3 = fix_inputs(x3, y3)  # type: ignore
    if not check_inputs(
            y1, y2, label1, label2):
        return
    if not check_inputs(
            x3, y3, xlabel, label3):
        return

    if len(x1) <= 1 or len(y1) <= 1 or len(y2) <= 1 or len(x2) <= 1:
        raise ValueError(
            f"The data for plot {filename} contains empty rows!")
    assert len(color) == 3

    fig, ax1 = plt.subplots()
    fig.subplots_adjust(right=0.75)
    # access colour
    prop_cycle = plt.rcParams["axes.prop_cycle"]
    try:
        linestyle = prop_cycle.by_key()["linestyle"]
    except KeyError:
        linestyle = rwth_cycle.by_key()["linestyle"]
    colors = prop_cycle.by_key()["color"]

    if max(color) >= len(colors):
        colors += colors
        linestyle += linestyle

    # first plot
    lines = ax1.plot(x1, y1, label=label1,
                     color=colors[color[0]],
                     linestyle=linestyle[0])
    ax1.set_xlabel(xlabel)
    ax1.set_ylabel(ylabel)
    ax1.set_ylim(
        min(y1) - (max(y1) - min(y1)) * 0.02,
        max(y1) + (max(y1) - min(y1)) * 0.02
    )
    ax1.yaxis.label.set_color(colors[color[0]])
    ax1.tick_params(axis="y", colors=colors[color[0]])

    # second plot
    ax2 = ax1.twinx()
    lines += ax2.plot(x2, y2, label=label2,
                      color=colors[color[1]],
                      linestyle=linestyle[1])
    ax2.set_ylabel(ylabel2)
    ax2.set_ylim(
        min(y2) - (max(y2) - min(y2)) * 0.02,
        max(y2) + (max(y2) - min(y2)) * 0.02
    )
    ax2.yaxis.label.set_color(colors[color[1]])
    ax2.tick_params(axis="y", colors=colors[color[1]])

    # third plot
    ax3 = ax1.twinx()
    make_invisible(ax3)
    ax3.spines["right"].set_position(("axes", 1.25))
    ax3.spines["right"].set_visible(True)
    lines += ax3.plot(x3, y3, label=label3,
                      color=colors[color[2]],
                      linestyle=linestyle[2])
    ax3.set_ylabel(ylabel3)
    ax3.set_ylim(
        min(y3) - (max(y3) - min(y3)) * 0.02,
        max(y3) + (max(y3) - min(y3)) * 0.02
    )
    ax3.yaxis.label.set_color(colors[color[2]])
    ax3.tick_params(axis="y", colors=colors[color[2]])

    # general settings
    if xlim is None:
        plt.xlim(min(x1), max(x1))
    else:
        plt.xlim(*xlim)
    labels = [line.get_label() for line in lines]
    if legend:
        plt.legend(lines, labels)
    # ticks
    if ticks is not None:
        ax2.set_yticks(ticks[0])
        ax2.set_yticklabels(ticks[1])
    plt.tight_layout()
    plt.savefig(filename)
    plt.close()



#!/usr/bin/env python3
"""
Stubs for mpl_toolskits' submodule mplot,
a library for additional function to be included with
matplotlib.
"""


class Axes:
    """Container for different axes."""

    class Axes3D:
        """
        3D-axes base class.
        """


class Mplot3D:
    """
    Class for the design of 3d surface plots.
    """

    axes3d: Axes



#!/usr/bin/env python
"""This contains the animation module of matplotlib."""
from typing import Callable

from .figure import Figure


class FuncAnimation:
    """Animation class for matplotlib-plots."""

    def __init__(self, fig: Figure,
                 update: Callable[[int], None],
                 frames: int = 10,
                 repeat: bool = False) -> None: ...



#!/usr/bin/env python3
"""
This file contains the settings for 2D surface
plots
"""
from __future__ import annotations

from os.path import join
from math import pi
from queue import Queue
from threading import Thread
from subprocess import check_output
from typing import (
    List, Tuple, TypeVar, Union, Iterable, Any, Optional)
from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
from numpy import array

from .plot_settings import apply_styles, rwth_gradient_map
from .types_ import Vector

mpl.use("Agg")
SURFACEFOLDER = join("simulation", "surface_data")

In = TypeVar("In", List[float], Tuple[float],
             Vector)


@apply_styles
def create_two_d_scatter_plot(
        X: In, Y: In, Z: In,
        folder: Union[str, Path],
        plot_title: str,
        xlabel: Optional[str], ylabel: Optional[str], zlabel: str)\
        -> None:
    """create two_d_plots"""
    # rearrange x, y, z to fit the shape needed for the plot
    fig = plt.figure()
    plt.set_cmap("jet")

    ax = fig.add_subplot(projection="3d")
    ax.scatter(Y, X, Z, cmap=rwth_gradient_map)

    if xlabel:
        ax.set_ylabel(xlabel)
    if ylabel:
        ax.set_xlabel(ylabel)
    ax.set_zlabel(zlabel)

    plt.tight_layout()

    plt.savefig(join(folder, plot_title.replace(" ", "_") + ".pdf"))


@apply_styles
def create_two_d_surface_plot(
        X: In, Y: In, Z: In,
        folder: Union[str, Path],
        plot_title: str,
        xlabel: Optional[str], ylabel: Optional[str], zlabel: str)\
        -> None:
    """create two_d_plots"""
    # rearrange x, y, z to fit the shape needed for the plot
    fig = plt.figure()
    plt.set_cmap("jet")
    Z_flat: Vector = array(Z)

    # X_two_d, Y_two_d=meshgrid(X_flat, Y_flat)

    ax = fig.add_subplot(projection="3d")
    # ax.plot_surface(X_two_d, Y_two_d, Z_flat, cmap=rwth_gradient_map)
    ax.plot_trisurf(Y, X, Z, cmap=rwth_gradient_map)

    if ylabel:
        ax.set_ylabel(ylabel)
    if xlabel:
        ax.set_xlabel(xlabel)
    ax.set_zlabel(zlabel)

    ax.set_zlim(min(Z_flat) * 0.98, max(Z_flat) * 1.05)
    ax.set_xlim(min(Y), max(Y))
    ax.set_ylim(min(X), max(X))

    plt.tight_layout()

    plt.savefig(join(folder, plot_title.replace(" ", "_") + ".pdf"))


def get_leakage(data: Iterable[Any], var: str = "density",
                surface_file: Optional[str] = None) -> list[float]:
    """calculate the leakage for a given set of data
    @param data enumerable set of valve-objects
        which allow the determination of the leakage
    @return list of the same dimension for the leakage"""
    if surface_file is None:
        surface_path = join(SURFACEFOLDER, "c_q.dat")
    leakage_bin = join(".", "subroutines", "bin", "test_leakage")
    Y: list[float] = []
    X: list[float] = []
    q: Queue[Any] = Queue()
    # run every call of the fortran-code in parallel
    for d in data:  # put the data into the
        # queue to access them later as needed
        q.put(d)

    def work_of_queue() -> None:
        nonlocal X
        nonlocal Y
        while True:
            d = q.get()
            if d is None:
                return  # last data-point
            pressure = max(d.short.p)
            print(pressure)
            print(d.angle, d.wobble)
            C = float(check_output([leakage_bin, "-i", surface_path, "-P",
                                    f"{pressure}"]))
            # A=d.short.unroundness2
            A = d.short.sigma
            R = d.valve.seat.radius
            delta_p = d.dictionary["fluid-pressure"]["value"]
            Y += [delta_p * 2.0 * pi * R / A * C]
            X += [getattr(d, var)]

    threads = [Thread(target=work_of_queue) for i in range(16)]
    for thread in threads:  # start all threads
        thread.start()
        q.put(None)
    for thread in threads:  # wait for all threads to finish
        thread.join()
    return Y


def plot_2d_surface(
    data: Iterable[Any],
    folder: str = "simulation",
    var1: str = "angle",
    var2: str = "wobble",
    xlabel1: Optional[str] = None,
    xlabel2: Optional[str] = None,
    surface_file: Optional[str] = None,
) -> None:
    """create the two d surface plots of two given variables"""
    X = [getattr(d, var1) for d in data]
    Y = [getattr(d, var2) for d in data]
    pressure = [max(d.short.p) for d in data]
    A = [d.short.unroundness for d in data]
    leakage = get_leakage(data, surface_file=surface_file)

    create_two_d_scatter_plot(
        X, Y, pressure, folder, "two d pressure",
        xlabel1, xlabel2, "maximal pressure [MPa]"
    )
    create_two_d_scatter_plot(
        X, Y, A, folder, "two d area", xlabel1, xlabel2, "contact area [mm]"
    )
    create_two_d_scatter_plot(
        X, Y, leakage, folder,
        "two d leakage", xlabel2, xlabel2, "leakage [ml/s]"
    )
    create_two_d_surface_plot(
        X,
        Y,
        pressure,
        folder,
        "two d pressure surface",
        xlabel1,
        xlabel2,
        "maximal pressure [MPa]",
    )
    create_two_d_surface_plot(
        X, Y, A, folder, "two d area surface",
        xlabel1, xlabel2, "contact area [mm]"
    )
    create_two_d_surface_plot(
        X,
        Y,
        pressure,
        folder,
        "two d pressure surface",
        xlabel1,
        xlabel2,
        "maximal pressure [MPa]",
    )
    create_two_d_surface_plot(
        X, Y, leakage, folder, "two d leakage surface",
        xlabel2, xlabel2, "leakage [ml/s]"
    )
